RWTexture2D<float4> tResult : register(u0);
RWTexture2D<float4> tGICache : register(u1);
RWTexture2D<float4> tGICachePrev : register(u2);
// CUSTOMIZABLE STUFF
#define GI_MAX_HIT_DIST 3.0f
#define REFL_MAX_HIT_DIST 100.0f
#define REFLECTIONS 1
#define GI 0
#define SHADOWS 0
#define LIGHT_SIZE 0.053f
#define REFLECTIONS_THRESHOLD 0.3f
#define ALPHA_THRESHOLD 0.3f

cbuffer SceneConstants : register( b0 )
{
    float4x4 mView;
    float4x4 mProjection;
    float4x4 mViewProjection;
    float4x4 mInvViewProjection;
    float4 vViewDir;
    float4 vViewPos;
    float4 deltas;
    float4 padd_[2];
};

cbuffer TimecycleConstants : register( b1 )
{
    float4 cTopSkyColor;
    float4 cBottomSkyColor;
    float4 vSunDir;
};

struct Triangle
{
    uint2 ids;
};

struct Material
{
    int textureId;
    uint flags;
};

struct AABB
{
    float4 aabb_min;
    float4 aabb_max;
};

struct BLASBVHLeaf
{
    AABB aabb;
    uint primitivesOffset;
    uint secondChildOffset;
    uint primitiveCount;
    uint axis;
};

struct TLASBVHLeaf
{
    AABB aabb;

    uint secondChildOffset;
    uint lowLevel;
    uint lowLevel_a;
    uint materialOffset;

    uint blasBVHOffset;
    uint primitiveOffset;
    uint vertexOffset;
    uint axis;

    float4x4 world_transform;
};

struct vertex
{
    float3 pos;
    uint color;
    float2 texcoords;
    float2 params;
};

StructuredBuffer<Triangle> bufTriList : register(t0);
StructuredBuffer<vertex> bufVertexList : register(t1);
StructuredBuffer<BLASBVHLeaf> bufAABBList : register(t2);
StructuredBuffer<TLASBVHLeaf> bufTLASbvh : register(t3);
Texture2D gb0 : register(t4);
Texture2D gb1 : register(t5);
Texture2D gb2 : register(t6);
Texture2D gb3 : register(t7);
Texture2D<float4> blue_noise : register(t8);
StructuredBuffer<Material> bufMaterials : register(t9);
Texture2DArray<float4> bufTextures : register(t10);
SamplerState sampl_st : register(s0);

struct Ray
{
    float4 origin;
    float4 direction;
    float4 invdirection;
};

struct PayLoad
{
    float3 normal;
    float hitDist;
    float4 color;
    float4 hitPosition;
};

uint get_tri_x( uint2 tri )
{
    return tri.x & 0xFFFF;
}
uint get_tri_y( uint2 tri )
{
    return (tri.x >> 16) & 0xFFFF;
}
uint get_tri_z( uint2 tri )
{
    return tri.y & 0xFFFF;
}
uint get_tri_matid( uint2 tri )
{
    return (tri.y >> 16) & 0xFFFF;
}

inline float GetOpacity( uint flags )
{
    uint opacity_ui = (flags)&0xFF;
    return float(opacity_ui)/255.0f;
}

inline uint GetMipLevel( uint flags )
{
    return (flags>>8)&0xFF;
}

inline float2 GetScale( uint flags )
{
    return float2( 1.0f/float((flags>>16)&0xFF), 1.0f/float((flags>>24)&0xFF) );
}

float4 unpackColor(uint color)
{
    float d = 1.0f/255.0f;
    return float4( float((color) & 0xFF)*d, float((color >> 8) & 0xFF)*d, float((color >> 16) & 0xFF)*d, float((color >> 24) & 0xFF)*d );
}

bool RayTriangleIntersection(Ray r, Triangle tri, out PayLoad payLoad, uint vertex_offset, uint material_offset, out bool is_alpha)
{
	is_alpha = false;
    vertex v0_vtx = bufVertexList[get_tri_x(tri.ids) + vertex_offset];
    vertex v1_vtx = bufVertexList[get_tri_y(tri.ids) + vertex_offset];
    vertex v2_vtx = bufVertexList[get_tri_z(tri.ids) + vertex_offset];
    float3 v0 = v0_vtx.pos.xyz;
    float3 v1 = v1_vtx.pos.xyz;
    float3 v2 = v2_vtx.pos.xyz;

    float4 c0 = unpackColor(v0_vtx.color);
    float4 c1 = unpackColor(v1_vtx.color);
    float4 c2 = unpackColor(v2_vtx.color);
    
    //t.color
    float3 e1 = v1 - v0;
    float3 e2 = v2 - v0;
    // normal calc
    float3 pvec = cross(r.direction.xyz, e2);
    float det = dot(e1, pvec);

    // Ray is parallel
    if (det < 1e-8 && det > -1e-8)
    {
        return false;
    }

    float inv_det = 1 / det;
    float3 tvec = r.origin.xyz - v0;
    float u = dot(tvec, pvec) * inv_det;
    if (u < 0 || u > 1)
    {
        return false;
    }

    float3 qvec = cross(tvec, e1);
    float v = dot(r.direction.xyz, qvec) * inv_det;
    if (v < 0 || u + v > 1)
    {
        return false;
    }
    payLoad.hitDist = dot(e2, qvec) * inv_det;
    payLoad.normal = normalize(cross(e1, e2));
    payLoad.color = pow(float4((1-u-v)*c0+u*c1+v*c2),2);
    payLoad.hitPosition = r.origin + r.direction * payLoad.hitDist;
	if (dot(r.direction,payLoad.normal) > 0) 
        return false;
    if (payLoad.hitDist < 0)
        return false;
    Material triangle_mat = bufMaterials[material_offset + get_tri_matid(tri.ids)];
    float opacity = GetOpacity(triangle_mat.flags);
    if( triangle_mat.textureId > 0 )
    {
        uint mip_level = GetMipLevel(triangle_mat.flags);
        uint size = 1024>>mip_level;
        float2 tc_scale = GetScale(triangle_mat.flags);
        float2 texcoord = v0_vtx.texcoords * (1-u-v) + u* v1_vtx.texcoords + v* v2_vtx.texcoords;
        float3 pos_xy_and_array = float3(texcoord.x*tc_scale.x, texcoord.y*tc_scale.y,float(triangle_mat.textureId));
        if( triangle_mat.textureId < 128 )
            payLoad.color = bufTextures.SampleLevel(sampl_st, pos_xy_and_array, mip_level);
		
		is_alpha = payLoad.color.a<=ALPHA_THRESHOLD;
        return !is_alpha; //bufTextures.mips[mip_level][pos_xy_and_array].a>0.3f;
    }
    else
        return true;
}


bool RayAABBIntersection(Ray r, float3 bbmin, float3 bbmax, out PayLoad payLoad)
{
    payLoad.hitDist = distance(r.origin.xyz, (bbmax + bbmin)*0.5f);
    payLoad.normal = float3(0,0,0);
    payLoad.color = float4(((bbmin + bbmax) * 0.5), 0.8);
    payLoad.hitPosition = r.origin + r.direction * payLoad.hitDist;
    /*
    float tmax=1000, tmin=0;
    for ( int i = 0; i < 3; i++ )
    {
        float invD = r.invdirection[i];
        float t0 = ( bbmin[i] - r.origin[i] ) * invD;
        float t1 = ( bbmax[i] - r.origin[i] ) * invD;
        if ( invD < 0.0f )
        {
            float temp = t0;
            t0 = t1;
            t1 = temp;
        }
        tmin = t0 > tmin ? t0 : tmin;
        tmax = t1 < tmax ? t1 : tmax;
        if( tmax < tmin )
            return false;
    }
    payLoad.hitDist = min(tmin, tmax);*/
    const float3 t0 = (bbmin - r.origin.xyz) * r.invdirection.xyz;
    const float3 t1 = (bbmax - r.origin.xyz) * r.invdirection.xyz;
 
    const float3 tmax = max(t0, t1);
    const float3 tmin = min(t0, t1);
 
    const float a1 = min(tmax.x, min(tmax.y, tmax.z));
    const float a0 = max(max(tmin.x, tmin.y), max(tmin.z, 0.0f));
 
    if (a1 < a0)
        return false;
    payLoad.hitDist = min(a0, a1);
    payLoad.hitPosition = r.origin + r.direction * payLoad.hitDist;
    return true;
}

float3 getSkyColor(Ray r){
    float dir = max(dot(r.direction, float3(0,0,1)),0.0f);
    return lerp(cBottomSkyColor, cTopSkyColor, dir ) + pow(max(dot(vSunDir,r.direction),0.0f),512.0f);
}

bool TraverseSceneBLAS(Ray r, TLASBVHLeaf tlasLeaf, out PayLoad payLoad, float maxDist, bool any_hit)
{
    r.origin = mul(tlasLeaf.world_transform,r.origin);
    r.origin.a = 1;
    r.direction = mul(tlasLeaf.world_transform, r.direction);
    r.direction.a = 0;
    r.invdirection =  float4( 1.0f / r.direction.x, 
                              1.0f / r.direction.y,
                              1.0f / r.direction.z, 1);

    payLoad.hitDist = maxDist;
    payLoad.color.rgb = getSkyColor(r);
    payLoad.color.a = 1;
    payLoad.normal = float3(0,0,0);

    PayLoad payLoadCurr;
    payLoadCurr.hitDist = payLoad.hitDist;
    
    uint todoOffset = 0, nodeNum = 0;
    int todo[64];
    bool hit = false;
    
    bool dirIsNeg[3] = { r.invdirection.x < 0, r.invdirection.y < 0, r.invdirection.z < 0 };
    while (true)
    {
        BLASBVHLeaf currentLeaf = bufAABBList[nodeNum + tlasLeaf.blasBVHOffset];
        if (RayAABBIntersection(r, currentLeaf.aabb.aabb_min, 
                                   currentLeaf.aabb.aabb_max, payLoadCurr))
        {
            if (currentLeaf.primitiveCount > 0)
            {
                for (int i = currentLeaf.primitivesOffset;
                         i < currentLeaf.primitivesOffset + currentLeaf.primitiveCount; 
                         i++)
                {
					bool is_alpha = false;
                    if ( RayTriangleIntersection( r, bufTriList[i + tlasLeaf.primitiveOffset], 
                                                 payLoadCurr, tlasLeaf.vertexOffset, tlasLeaf.materialOffset, is_alpha ) )
                    {
                        if ( payLoadCurr.hitDist < payLoad.hitDist && 
                             payLoadCurr.hitDist > 0.0f )
                        {
                            payLoad = payLoadCurr;
                        }
                        hit = true;
                        if(any_hit)
                            break;
                    }
                }
                if (todoOffset == 0) break;
                nodeNum = todo[--todoOffset];
            }
            else
            {
                if(dirIsNeg[currentLeaf.axis])
                {
                    todo[todoOffset++] = nodeNum + 1;
                    nodeNum = currentLeaf.secondChildOffset;
                }
                else{
                    todo[todoOffset++] = currentLeaf.secondChildOffset;
                    nodeNum = nodeNum + 1;
                }
            }
        }
        else
        {
            if (todoOffset == 0) break;
            nodeNum = todo[--todoOffset];
        }
    }
    if(hit)
    {
        payLoad.normal = mul(float4(payLoad.normal,0),tlasLeaf.world_transform);
    }
    return hit;
}

bool TraverseSceneTLAS(Ray r, out PayLoad payLoad, float maxDist, bool any_hit)
{
    payLoad.hitDist = maxDist;
    payLoad.color.rgb = getSkyColor(r);
    payLoad.color.a = 1;
    payLoad.normal = float3(0,0,0);

    PayLoad payLoadCurr;
    payLoadCurr.hitDist = payLoad.hitDist;
    

    int currentDepth = 0;
    int maxDepth = 1000;

    uint todoOffset = 0, nodeNum = 0;
    int todo[64];
    bool hit = false;
    
    bool dirIsNeg[3] = { r.invdirection.x < 0, r.invdirection.y < 0, r.invdirection.z < 0 };
    while (true)
    {
        TLASBVHLeaf currentLeaf = bufTLASbvh[nodeNum];
        if (RayAABBIntersection(r, currentLeaf.aabb.aabb_min, 
                                   currentLeaf.aabb.aabb_max, payLoadCurr) && payLoadCurr.hitDist<maxDist)
        {
            if (currentLeaf.lowLevel.x > 0)
            {
                if (TraverseSceneBLAS(r, currentLeaf, payLoadCurr, maxDist,any_hit))
                {
                    if (payLoadCurr.hitDist < payLoad.hitDist && payLoadCurr.hitDist > 0.0f)
                    {
                        payLoad = payLoadCurr;
                    }
                    hit = true;
                    if(any_hit)
                        break;
                }
                if (todoOffset == 0) break;
                nodeNum = todo[--todoOffset];
            }
            else
            {
                if(dirIsNeg[currentLeaf.axis])
                {
                    todo[todoOffset++] = nodeNum + 1;
                    nodeNum = currentLeaf.secondChildOffset;
                }
                else{
                    todo[todoOffset++] = currentLeaf.secondChildOffset;
                    nodeNum = nodeNum + 1;
                }
            }
        }
        else
        {
            if (todoOffset == 0) break;
            nodeNum = todo[--todoOffset];
        }
    }
    return hit;
}

float rand_1_05(in float2 uv)
{
    float2 noise = (frac(sin(dot(uv, float2(12.9898, 78.233) * 2.0)) * 43758.5453));
    return abs(noise.x + noise.y) * 0.5;
}
float rand_1_06(in float2 uv)
{
    float2 noise = (frac(sin(dot(uv, float2(22.9898, 32.233) * 2.0)) * 64564.5453));
    return abs(noise.x + noise.y) * 0.5;
}
float rand_1_07(in float2 uv)
{
    float2 noise = (frac(sin(dot(uv, float2(56.9898, 85.233) * 2.0)) * 12346.5453));
    return abs(noise.x + noise.y) * 0.5;
}
float3 randomPointOnUnitSphere(uint2 seed)
{
    float3 randomPointOnSphere;
    float3 pseudo_random_lut = normalize( blue_noise[uint2((seed.x)%16,(seed.y)%16)].xyz ) * 2.0 - 1.0;
    
    //pseudo_random_lut = frac(pseudo_random_lut + deltas.xyy);

    float theta = 2 * 3.14f * pseudo_random_lut.y;
    float phi = acos(1 - pseudo_random_lut.z*2);
    float x = sin(phi) * cos(theta);
    float y = sin(phi) * sin(theta);
    float z = cos(phi);
    randomPointOnSphere =pseudo_random_lut; //normalize(float3(x,y,z));
    return randomPointOnSphere;
}

float3 PointOnHeminormal(float3 normal, float3 pt)
{
	return float3(pt.x > normal.x ? pt.x : -pt.x, pt.y > normal.y ? pt.y : -pt.y, pt.z > normal.z ? pt.z : -pt.z );
}

inline Ray GenerateCameraRay(uint2 index, in float3 cameraPosition, in float4x4 cameraMatrix)
{
    
    float screen_width = 1600.0f;
    float screen_height = 900.0f;
    float aspect_ratio = screen_width/screen_height;

    float2 xy = index; // center in the middle of the pixel.
    float2 screenPos = float2(xy.x/(screen_width),xy.y/(screen_height)) * 2.0 - 1.0;

    // Invert Y for DirectX-style coordinates.
    screenPos.y = -screenPos.y;

    float4 camWorldDir = mul(cameraMatrix, float4( screenPos.x / mProjection[0].x, screenPos.y/mProjection[1].y, 1, 0) );

    Ray ray;
    ray.origin = float4( cameraPosition, 1 );
    ray.direction =float4( normalize(camWorldDir.xyz), 0);
    ray.invdirection = float4( 1.0f / ray.direction.x, 
                               1.0f / ray.direction.y, 
                               1.0f / ray.direction.z, 0);

    return ray;
}

#define IRRADIANCE_CACHE_A 64
#define IRRADIANCE_CACHE_LIMIT ((1280*720)/(IRRADIANCE_CACHE_A*IRRADIANCE_CACHE_A))

void SaveToCache( float3 pt, float hdist, float3 n, float3 color, uint id ){
    //160*90

    // x x x x
    // y y y y
    // z z z z

    // x x x x
    // y y y y
    // z z z z

    uint2 pos_id = uint2(id%160, (id/160)*3);
    uint2 norm_id = pos_id + uint2(0, 1);
    uint2 color_id = norm_id + uint2(0, 1);
    tGICache[pos_id] = float4( pt, 6.0f/hdist );
    tGICache[norm_id] = float4( n, 1.0f );
    tGICache[color_id] = float4( tGICachePrev[ color_id ].xyz *5.0f/6.0f + color/6.0f , 1.0f );
}

bool SearchValidIds( float3 pt, out uint last_id )
{
    uint id = 0;
    uint2 pos_id = uint2(id%160, (id/160)*3);
    uint2 norm_id = pos_id + uint2(0, 1);

    while( tGICache[ norm_id ].w > 0 && id < IRRADIANCE_CACHE_LIMIT)
    {
        if( distance( tGICache[ pos_id ].xyz, pt ) < tGICache[ pos_id ].w )
        {
            return true;
        }
        id++;
        pos_id = uint2(id%160, (id/160)*3);
        norm_id = pos_id + uint2(0, 1);
    }
    last_id = id;
    if(last_id == IRRADIANCE_CACHE_LIMIT-1)
        id = 1;
    else
        id += 1;
    
    pos_id = uint2(id%160, (id/160)*3);
    norm_id = pos_id + uint2(0, 1);
    tGICachePrev[ norm_id ]=float4(0,0,0,0);
    return false;
}

float4 GetCachedGI( float3 pt, float3 n, uint id ){
    //160*90
    // x x x x
    // x x x x
    uint2 pos_id = uint2(id%160, (id/160)*3);
    uint2 norm_id = pos_id + uint2(0, 1);
    uint2 color_id = norm_id + uint2(0, 1);
    float w = distance( pt , tGICachePrev[ pos_id ].xyz ) / tGICachePrev[ pos_id ].w +
              sqrt(1 + dot(tGICachePrev[ norm_id ].xyz, n));
    if(w > 0)
        return float4(tGICachePrev[ color_id ].xyz/w, 1.0f/w);
    else
        return float4(0.0f,0.0f,0.0f,0.0f);
}

inline float pow5(float v) {
	return v*v*v*v*v;
}

float MicrofacetFresnel(in float3 LightDir, in float3 Normal, in float fRoughness) {
	float IOR = 1.5f;
    float f0 = (1 - IOR) / (1 + IOR);
	f0 *= f0;
    // Cosine between light and normal
    float CosPhi = max(dot(LightDir, Normal), 0);

    return f0 + (1 - f0) * pow5(1 - CosPhi);
}

[numthreads(8, 8, 1)]
void TraceRays(uint3 DTid : SV_DispatchThreadID, uint groupid : SV_GroupIndex)
{
    float3 light = vSunDir;
    float lightSize = LIGHT_SIZE;

    float3 lowerLeft = float3(-0.5f, 0.5f, 1.0f);
    float3 u = float3(1.33f, 0.0f, 0.0f);
    float3 v = float3(0.0f, -1.0f, 0.0f);
    float screen_width = 1600.0f;
    float screen_height = 900.0f;
	
    Ray r=GenerateCameraRay(DTid.xy,vViewPos.xyz,mInvViewProjection);

    float dist;
    float eps = 0.001f;

    PayLoad payLoad;

    float4 resultColor = float4(getSkyColor(r), 1);
    float4 texture_color = gb2.SampleLevel(sampl_st, float2(DTid.x/screen_width,DTid.y/screen_height),0);
    float auto_reflectivity = min((dot(texture_color.xyz,texture_color.xyz))*0.7,0.7);
    if( texture_color.a < 0.3f )
    {
        tResult[DTid.xy] = resultColor;
        return;
    }
    //TraverseSceneTLAS(r,payLoad,900, false);
    //tResult[DTid.xy] = payLoad.color;
    //return;
    //if(TraverseSceneTLAS(r,payLoad,900))
    //{
        float4 hitPoint = gb0.SampleLevel(sampl_st, float2(DTid.x/screen_width,DTid.y/screen_height),0);
        float3 hpNormal;
        float4 n_c_gb = gb1.SampleLevel(sampl_st, float2(DTid.x/screen_width,DTid.y/screen_height),0);
        hpNormal = n_c_gb.xyz;
        float4 rad_color = gb3.SampleLevel(sampl_st, float2(DTid.x/screen_width,DTid.y/screen_height),0);
        //r.origin + r.direction * payLoad.hitDist;
        Ray shadow_ray;
        shadow_ray.origin = hitPoint + 
                            float4( hpNormal * eps, 0 );
        shadow_ray.origin.a = 1;
        shadow_ray.direction = float4( normalize(light + randomPointOnUnitSphere(DTid.xy)*lightSize),0);
        shadow_ray.invdirection = float4( 1.0f / shadow_ray.direction.x, 
                                          1.0f / shadow_ray.direction.y,
                                          1.0f / shadow_ray.direction.z, 1);
        
        
        PayLoad shadow_pay_load;
        float ndl = saturate( max(dot(normalize(light),normalize(hpNormal)),0.0) );
        bool has_shadows = true;
#if SHADOWS == 1
            has_shadows = TraverseSceneTLAS(shadow_ray,shadow_pay_load,1000,true);
#endif
        resultColor = float4(ndl.xxx*!has_shadows,1);
        
        Ray gi_ray;
        gi_ray.origin = hitPoint + 
                                float4( hpNormal * eps, 0 );
        gi_ray.direction = float4( normalize(randomPointOnUnitSphere(DTid.xy) + hpNormal),0);
        gi_ray.invdirection = float4( 1.0f / gi_ray.direction.x, 
                                      1.0f / gi_ray.direction.y,
                                      1.0f / gi_ray.direction.z, 1);
        uint s = 3;
        uint2 dxCyC =(DTid.xy)/8;
        uint cache_id = 0;
        float4 gi = float4(0.0,0.0,0.0,0);
#if GI == 1
        if(/*!SearchValidIds(hitPoint.xyz,cache_id)*/ true)
        {
            PayLoad gi_load;
            bool gi_hit = TraverseSceneTLAS(gi_ray,gi_load,GI_MAX_HIT_DIST,false);

            gi.xyz = gi_load.color.xyz;
            gi.a = 1;
            
            if( gi_hit )
            {
                PayLoad gi_spayLoad;
                
                Ray gi_s_ray;
                gi_s_ray.origin = hitPoint + gi_ray.direction * gi_load.hitDist + 
                                float4( gi_load.normal * eps, 0 );
                gi_s_ray.origin.a = 1;
                gi_s_ray.direction = float4( normalize(light),0);
                gi_s_ray.invdirection = float4( 1.0f / gi_s_ray.direction.x, 
                                                1.0f / gi_s_ray.direction.y,
                                                1.0f / gi_s_ray.direction.z, 0);
                bool gi_shadows = TraverseSceneTLAS(gi_s_ray,gi_spayLoad,1000,true);
                if( gi_shadows )
                    gi.xyz = 0.0f;
            }
            float mhdst = GI_MAX_HIT_DIST;

            /*for(int s = 0; s< 1; s++)
            {
                gi_ray.origin += gi_ray.direction * gi_load.hitDist + 
                                        float4( gi_load.normal.xyz * eps, 0 );
                gi_ray.direction = float4( normalize(randomPointOnUnitSphere(DTid.xy) + gi_load.normal.xyz),0);
                gi_ray.invdirection = float4( 1.0f / gi_ray.direction.x, 
                                            1.0f / gi_ray.direction.y,
                                            1.0f / gi_ray.direction.z, 1);
                TraverseSceneTLAS(gi_ray,gi_load,mhdst,false);
                
                gi.xyz = gi_load.color.xyz + max(dot( gi_ray.direction.xyz, gi_load.normal.xyz), 0.0f) * gi.xyz;
                gi.a = 1;
            }*/
            
            //SaveToCache( hitPoint.xyz, gi_load.hitDist, hpNormal.xyz, gi.xyz, 
            //            (DTid.x / IRRADIANCE_CACHE_A) * 30 + DTid.y / IRRADIANCE_CACHE_A);
        }
        //resultColor.xyz += rad_color.rgb*0.5f;
#else
        //resultColor.xyz += rad_color.rgb*0.5f;
#endif
       // for(uint id = 0; id < IRRADIANCE_CACHE_LIMIT; id++)
       // {
       //     gi += GetCachedGI(hitPoint.xyz,hpNormal.xyz,id);
       // }
        resultColor.xyz += gi.xyz;
        //resultColor.xyz +=rad_color.xyz;
#if REFLECTIONS == 1
		float3 viewDir = normalize(hitPoint.xyz - vViewPos.xyz);
        float fresnel = MicrofacetFresnel(normalize(-viewDir.xyz), normalize(hpNormal), 1-auto_reflectivity);
        Ray refl_ray;
        refl_ray.origin = hitPoint + float4( hpNormal * eps, 0 );
        refl_ray.direction = float4( normalize( normalize(reflect( viewDir, hpNormal.xyz )) + randomPointOnUnitSphere(DTid.xy) * (1-auto_reflectivity) * 0.015 )
										,0);
        refl_ray.invdirection = float4( 1.0f / refl_ray.direction.x, 
                                      1.0f / refl_ray.direction.y,
                                      1.0f / refl_ray.direction.z, 1);
            
        PayLoad refl_load;
        Ray refr_ray;
        refr_ray.origin = hitPoint + 
                                float4( hpNormal.xyz * eps, 0 );
        refr_ray.direction = float4( normalize(r.direction.xyz),0);
        refr_ray.invdirection = float4( 1.0f / refr_ray.direction.x, 
                                      1.0f / refr_ray.direction.y,
                                      1.0f / refr_ray.direction.z, 1);
            
        PayLoad refr_load;
		//if(auto_reflectivity < REFLECTIONS_THRESHOLD)
        {
			TraverseSceneTLAS(refl_ray,refl_load,(REFL_MAX_HIT_DIST),false);
			resultColor.xyz += refl_load.color.rgb;
		}
        //TraverseSceneTLAS(refr_ray,refr_load,100.0,false);
            //resultColor.xyz = refl_load.color.rgb * fresnel + refr_load.color.rgb * (1-fresnel);

#endif
        //resultColor.xyz *= payLoad.color.xyz;
        //resultColor.xyz = payLoad.normal.xyz;
        //resultColor.xyz += refl_load.color*0.5f;
    //}

    //if (TraverseSceneBVH(r, payLoad, 1000))

    tResult[DTid.xy] = resultColor;//+ (max_r + max_l)
                          //  
                           //)/3.0f;
}